"""This module allows getting information and changing Alma holding records"""

from typing import Optional, Literal, List, ClassVar, Union
import logging
from ..record import Record, check_error, XmlData
import almapiwrapper.inventory as inventory
from lxml import etree
import os


class Holding(Record):
    """
    Class representing a holding object. Holdings are only in the IZ.

    Several possibilities for building holdings:
    - 'get_holdings' method of IzBib object
    - Holding(mms_id, holding_id, zone, env)
    - Holding(bib=BibIz, holding_id=holding_id)

    if no 'holding_id' is provided, but 'data' is provided and create_holding is True, then
    it creates a new holding.


    :ivar holding_id: holding ID
    :ivar zone: zone of the record
    :ivar env: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar bib: class:`almapiwrapper.inventory.IzBib` object
    :ivar data: :class:`almapiwrapper.record.XmlData` or `etree.Element` object,
    useful to force update a record from a backup
    :param mms_id: record mms_id
    :param create_holding: boolean, if True try to create a new holding (if no 'holding_id' is provided)
    """

    api_base_url_bibs: ClassVar[str] = f'{Record.api_base_url}/bibs'

    def __init__(self,
                 mms_id: Optional[str] = None,
                 holding_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 bib: Optional[inventory.IzBib] = None,
                 data: Optional[Union['XmlData', etree.Element]] = None,
                 create_holding: Optional[bool] = False) -> None:
        """
        Construct a holding record. Several possibilities for building holdings:
        - 'get_holdings' method of IzBib object
        - Holding(mms_id, holding_id, zone, env)
        - Holding(bib=BibIz, holding_id=holding_id)

        if no 'holding_id' is provided, but 'data' is provided and create_holding is True, then
        it creates a new holding.
        """
        self._items = None
        self.error = False
        self.error_msg = None
        self._data = None
        self.area = 'Bibs'
        self.format = 'xml'
        self.bib = bib
        self.holding_id = holding_id

        if self.bib is not None:
            self.zone = self.bib.zone
            self.env = self.bib.env

        else:
            self.zone = zone
            self.env = env
            self.bib = inventory.IzBib(mms_id, self.zone, self.env)

            # If there is an error in the bibliographic record, it is spread out in the holding record
            if self.bib.error is True:
                self.error = True
        # Create a new holding if 'create_holding' is True
        if self.holding_id is None and data is not None and create_holding is True:
            if data.__class__.__name__ == '_Element':
                data = XmlData(etree.tostring(data))
            self._create_holding(data)

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.bib.mms_id}', '{self.holding_id}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[XmlData]:
        """Fetch holding data via API. Store the data in the 'data' attribute.

        :return: None
        """
        r = self._api_call('get',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding_id}',
                           headers=self._get_headers())

        if r.ok is True:
            logging.info(f'{repr(self)}: holding data available')
            return XmlData(r.content)
        else:
            self._handle_error(r, 'unable to fetch holding data')

    def _create_holding(self, data: etree.Element) -> None:
        """Create a holding and link it to the provided bibliographic record

        :param data: data used to create the holding

        :return: None
        """
        r = self._api_call('post',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings',
                           headers=self._get_headers(),
                           data=bytes(data))

        if r.ok is True:
            self.data = XmlData(r.content)
            self.holding_id = self.get_holding_id()
            logging.info(f'{repr(self)}: holding created')
        else:
            self._handle_error(r, 'unable to create the new holding')

    @check_error
    def get_holding_id(self) -> str:
        """get_holding_id(self) -> str
        Fetch holding ID in the data of 'data' attribute.

        Useful for creating a new holding.

        :return: str with holding id
        """
        return self.data.find('.//holding_id').text

    @check_error
    def save(self) -> 'Holding':
        """save(self) -> 'Holding'
        Save holding in a folder for each MMS ID.
        Example: records/UBS_9963486250105504/hol_22314215780005504_01.xml

        :return: Holding
        """
        filepath = f'records/{self.zone}_{self.bib.mms_id}/hol_{self.holding_id}.xml'
        self._save_from_path(filepath)
        return self

    @check_error
    def update(self) -> 'Holding':
        """update(self) -> 'Holding'
        Update data of a holding.

        :return: Holding
        """
        r = self._api_call('put',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding_id}',
                           data=etree.tostring(self.data),
                           headers=self._get_headers())

        if r.ok is True:
            self.data = XmlData(r.content)
            logging.info(f'{repr(self)}: holding data updated')
        else:
            self._handle_error(r, 'unable to update holding data')

        return self

    @check_error
    def delete(self, force: Optional[bool] = False) -> None:
        """delete(self, force: Optional[bool] = False) -> None
        Delete holding

        :return: None
        """
        if force is True:
            self.delete_items()

        r = self._api_call('delete',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)} deleted')
        else:
            self._handle_error(r, 'unable to delete the holding')

    @check_error
    def get_items(self) -> List['inventory.Item']:
        """get_items(self) -> List['inventory.Item']
        This method is used to retrieve the list of items and loads
        the xml data of these items. To avoid reloading this information,
        the items references are stored in the private attribute '_items'.

        :return: list of :class:`almapiwrapper.bib.Item` objects
        """
        # Check if items have been already fetched
        if self._items is not None:
            return self._items

        # Fetch the item's data through apis
        r = self._api_call('get',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding_id}/items',
                           params={'limit': '100'},
                           headers=self._get_headers())

        if r.ok is True:
            root = etree.fromstring(r.content, parser=self.parser)
        else:
            self._handle_error(r, 'unable to fetch items')
            return []

        items_data = root.findall('.//item_data')
        if len(items_data) > 99:
            logging.warning(f'{repr(self)}: at least 100 items, risk of missing items')

        # No item found
        if len(items_data) == 0:
            logging.warning(f'{repr(self)}: no item found')
            self._items = []
            return self._items

        logging.info(f'{repr(self)}: {len(items_data)} items fetched')

        # Fetch items one by one
        self._items = []
        for item in items_data:
            item_id = item.find('.//pid').text
            self._items.append(inventory.Item(holding=self, item_id=item_id))

        return self._items

    @check_error
    def delete_items(self) -> None:
        """delete_items(self) -> None
        Delete all items of the holding

        :return: None
        """
        for item in self.get_items():
            item.delete()

    @property
    @check_error
    def library(self) -> Optional[str]:
        """library(self) -> Optional[str]
        Property of the holding returning the library code

        :return: str containing library code
        """
        library = self.data.find('.//datafield[@tag="852"]/subfield[@code="b"]')
        if library is None:
            logging.warning(f'{repr(self)}: no library in the holding')
            return
        return library.text

    @library.setter
    @check_error
    def library(self, library_code: str) -> None:
        """library(self, library_code: str) -> None
        This setter is able to update the 852$b of the holding. But the field should already exist.

        :param library_code: code of the library to insert in 852$b field

        :return: None
        """
        library = self.data.find('.//datafield[@tag="852"]/subfield[@code="b"]')
        if library is None:
            logging.error(f'{repr(self)}: no library field in the holding -> not possible to update it')
            self.error = True
        else:
            logging.info(f'{repr(self)}: library changed from "{library.text}" to "{library_code}"')
            library.text = library_code

    @property
    @check_error
    def location(self) -> Optional[str]:
        """location(self) -> Optional[str]
        Property of the holding returning the library code

        :return: str containing library code
        """
        location = self.data.find('.//datafield[@tag="852"]/subfield[@code="c"]')
        if location is None:
            logging.warning(f'{repr(self)}: no location in the holding')
            return
        return location.text

    @location.setter
    @check_error
    def location(self, location_code: str) -> None:
        """location(self, location_code: str) -> None
        This setter is able to update the 852$c of the holding. But the field should already exist.

        :param location_code: location code to set to the holding

        :return: None
        """
        location = self.data.find('.//datafield[@tag="852"]/subfield[@code="c"]')
        if location is None:
            logging.error(f'{repr(self)}: no location field in the holding -> not possible to update it')
        else:
            logging.info(f'{repr(self)}: location changed from "{location.text}" to "{location_code}"')
            location.text = location_code

    @property
    @check_error
    def callnumber(self) -> Optional[str]:
        """callnumber(self) -> Optional[str]
        Property of the holding returning the callnumber

        :return: str containing callnumber
        """
        field852 = self.data.find('.//datafield[@tag="852"]')
        if field852 is None:
            return None

        callnumber_field = field852.find('./subfield[@code="j"]')

        if callnumber_field is None:
            callnumber_field = field852.find('./subfield[@code="h"]')

        if callnumber_field is None:
            logging.warning(f'{repr(self)}: no callnumber field in the holding')
            return None

        return callnumber_field.text

    @callnumber.setter
    @check_error
    def callnumber(self, callnumber_txt: str) -> None:
        """callnumber(self, callnumber_txt: str) -> None
        This setter is able to update the 852$j or 852$h of the holding. But the field should already exist.

        :param callnumber_txt: text of the callnumber to be set in j or h subfield

        :return: None
        """
        field852 = self.data.find('.//datafield[@tag="852"]')
        if field852 is None:
            return

        callnumber_field = field852.find('./subfield[@code="j"]')

        if callnumber_field is None:
            callnumber_field = field852.find('./subfield[@code="h"]')

        if callnumber_field is None:
            logging.error(f'{repr(self)}: no callnumber field in the holding -> not possible to update it')
            return

        logging.info(f'{repr(self)}: callnumber changed from "{callnumber_field.text}" to "{callnumber_txt}"')
        callnumber_field.text = callnumber_txt

    @staticmethod
    def get_data_from_disk(mms_id: str, holding_id: str, zone: str) -> Optional[XmlData]:
        """get_data_from_disk(mms_id, holding_id, zone)
        Fetch the data of the described record

        :param mms_id: bib record mms_id
        :param holding_id: holding record ID
        :param zone: zone of the record

        :return: :class:`almapiwrapper.record.XmlData` or None
        """
        if os.path.isdir(f'records/{zone}_{mms_id}') is False:
            return

        # Fetch all available filenames related to this record
        file_names = sorted([file_name for file_name in os.listdir(f'records/{zone}_{mms_id}')
                             if file_name.startswith(f'hol_{holding_id}') is True])

        if len(file_names) == 0:
            return

        return XmlData(filepath=f'records/{zone}_{mms_id}/{file_names[-1]}')



import abc
from .apikeys import ApiKeys
from typing import Optional, Callable, ClassVar, Literal, Dict, Union
import logging
import json
from lxml import etree
import time
import os
import requests


def check_error(fn: Callable) -> Callable:
    """Prevents operation if the record is containing an error

    :param fn: method that should not to be executed in case of error
    :return: wrapper function
    """
    def wrapper(*args, **kwargs):
        """Wrapper function
        """
        rec = args[0]
        if rec.error is False and rec.data is not None:
            rec = fn(*args, **kwargs)
        else:
            logging.error(f'{repr(rec)}: due to error to the record, process "{fn.__name__}" skipped.')
        return rec

    wrapper.__doc__ = fn.__doc__

    return wrapper


class Record(metaclass=abc.ABCMeta):
    """
    This class is representing entities like bibliographic records, items or holdings.
    It contains common methods to all entities.

    :ivar zone: initial value: zone of the record (can be "NZ")
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: initial value: :class:`almapiwrapper.record.XmlData`
    """
    # Used to get API keys
    k: ClassVar[ApiKeys] = ApiKeys()

    # Api urls
    api_base_url: ClassVar[str] = 'https://api-eu.hosted.exlibrisgroup.com/almaws/v1'

    # Using this parser avoids some problems with 'tostring' method
    parser: ClassVar[etree.XMLParser] = etree.XMLParser(remove_blank_text=True)

    @abc.abstractmethod
    def __init__(self,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 data: Optional[Union['JsonData', 'XmlData']] = None) -> None:
        """Abstract constructor, all entities have at least a zone and an environment.
        """
        self.error = False
        self.error_msg = None
        self.zone = zone
        self.env = env
        self.area = None
        self.format = None
        self._data = data

    @abc.abstractmethod
    def _fetch_data(self) -> None:
        """Abstract method. This method is different according to the type of record"""
        return None

    @staticmethod
    def _api_call(method: Literal['get', 'put', 'post', 'delete'], *args, **kwargs) -> Optional[requests.Response]:
        """Static method to handle http errors. Quit the program after 3 failed tries

        :param method: 'get', 'put', 'post' or 'delete' according to the api method call
        """

        for api_try in [1, 2, 3]:
            try:
                time.sleep(0.04)
                if method == 'get':
                    r = requests.get(*args, **kwargs)
                elif method == 'put':
                    r = requests.put(*args, **kwargs)
                elif method == 'post':
                    r = requests.post(*args, **kwargs)
                elif method == 'delete':
                    r = requests.delete(*args, **kwargs)
                else:
                    return None

                if 'X-Exl-Api-Remaining' in r.headers and int(r.headers['X-Exl-Api-Remaining']) < 5000:
                    logging.critical(f'Limit of the number of requests allowed critical - '
                                     f'{r.headers["X-Exl-Api-Remaining"]} reamaining => exiting of the program')
                    exit()

                # Retry later if threshold of 25 requests per second exceeded
                if r.status_code != 429:
                    return r

            except requests.exceptions.RequestException as err:
                logging.error(f'HTTP error: try {api_try} - message: {str(err)}')
            if api_try == 3:
                logging.critical(f'HTTP error: try {api_try} => exiting of the program')
                exit()
            time.sleep(3)

    @staticmethod
    def build_headers(data_format: Literal['json', 'xml'],
                      zone: str,
                      area: str,
                      rights: Literal['R', 'RW'] = 'RW',
                      env: Optional[Literal['P', 'S']] = 'P') -> Dict:
        """Build the headers for the API calls.

        :param data_format: "json" or "xml"
        :param zone: optional, if indicated allow to make the query in an other IZ
        :param area: area of the record, bibs, users for example
        :param rights: "R" for read only or "RW" for write and read rights
        :param env: environment of the api call: 'P' for production, 'S' for sandbox
        :return: dict with the headers
        """

        # Build header dict
        return {'content-type': f'application/{data_format}',
                'accept': f'application/{data_format}',
                'Authorization': 'apikey ' + ApiKeys().get_key(zone, area, rights, env)}

    @property
    def data(self) -> Optional[Union[Dict, etree.Element]]:
        """Property that get xml data with API call. If not available, make an api call

        :return: xml data
        """
        if self._data is None and self.error is False:
            self.data = self._fetch_data()

        if self._data is not None:
            return self._data.content

    @data.setter
    def data(self, data: Union['JsonData', 'XmlData']) -> None:
        """Property used to set xml data of an holding
        :param data: dictionary with holding xml data
        :return: None
        """
        self._data = data

    def __str__(self) -> str:
        """Return content of data attribute as a string
        :return: string data
        """
        if self.data is not None:
            return str(self._data)
        else:
            return ''

    def __bytes__(self):
        """Return content of data attribute as bytes
        :return: bytes data
        """
        if self.data is not None:
            return bytes(self._data)
        else:
            return b''

    def _save_from_path(self, filepath: str) -> None:
        """Save a record in the 'records' folder. Versioning is supported. A suffix is added to the file path.

        :param filepath: initial path of the saved record
        :return: None
        """
        # Fetch directory
        directory = os.path.dirname(filepath)

        # Create the directory if it doesn't exist
        os.makedirs(directory, exist_ok=True)

        # Add a suffix _xx to the file name before extension
        filename = os.path.basename(filepath)
        base_filename, extension = filename.rsplit('.', 1)
        files = os.listdir(directory)
        version = str(len([f for f in files if base_filename in f]) + 1).zfill(2)

        # Build the final path
        final_path = os.path.join(directory, f'{base_filename}_{version}.{extension}')

        # Write the file
        with open(final_path, 'w') as f:
            f.write(str(self))

        logging.info(f'{repr(self)}: record saved: {final_path}')

    def _get_headers(self,
                     data_format: Optional[Literal['json', 'xml']] = None,
                     zone: Optional[str] = None,
                     area: Optional[str] = None,
                     rights: Literal['R', 'RW'] = 'RW',
                     env: Optional[Literal['P', 'S']] = None) -> Dict:
        """Build the headers for the API calls

        :param data_format: "json" or "xml" according to the format of the record
        :param zone: optional, if indicated allow to make the query in another IZ
        que celle de l'objet courant
        :param env: environment of the api call: 'P' for production, 'S' for sandbox
        :return: dict with the header
        """

        # If values are not provided, fetch values of the current entity

        if zone is None:
            zone = self.zone
        if env is None:
            env = self.env
        if area is None:
            area = self.area
        if data_format is None:
            data_format = self.format

        return self.build_headers(data_format, zone, area, rights, env)

    def _handle_error(self, r: requests.models.Response, msg: str):
        """Set the record error attribute to True and write the logs about the error

        :param r: request response of the api
        :param msg: context message of the error
        :return: None
        """
        if 'json' in r.headers['Content-Type']:
            json_data = r.json()
            try:
                error_message = json_data['errorList']['error'][0]['errorMessage']
            except KeyError:
                error_message = 'unknown error'
        else:
            try:
                xml = etree.fromstring(r.content, parser=self.parser)
                error_message = xml.find('.//{http://com/exlibris/urm/general/xmlbeans}errorMessage').text
            except etree.XMLSyntaxError:
                error_message = 'unknown error'
        logging.error(f'{repr(self)} - {r.status_code if r is not None else "unknown"}: '
                      f'{msg} / {error_message}')
        self.error = True
        self.error_msg = error_message


class XmlData:
    """Simple class representing XML data

    Only very general methods are described here.
    Method to create, update or delete the data are at the record level.

    :ivar content: bytes with the xml data
    """
    # Using this parser avoids some problems with 'tostring' method
    parser: ClassVar[etree.XMLParser] = etree.XMLParser(remove_blank_text=True)

    def __init__(self, content: Optional[bytes] = None, filepath: Optional[str] = None):
        """Constructor of XmlData object"""

        if filepath is not None:
            content = self._read_file(filepath)

        if content is not None:
            self.content = etree.fromstring(content, parser=self.parser)
        else:
            self.content = None

    def __str__(self):
        return etree.tostring(self.content, pretty_print=True).decode()

    def __bytes__(self):
        return etree.tostring(self.content)

    def sort_fields(self) -> None:
        """Sort all the fields and subfields of the record.

        :return: None
        """
        # Check if at least one datafield is available
        record = self.content.find('.//record')
        if record is None:
            # No record found
            logging.error(f'{repr(self)}: sorting fields failed, no record available in data attribute')

        # Sort datafields and controlfields according to the tag attribute
        record[:] = sorted(record, key=lambda field_or_contr: field_or_contr.get('tag', '000'))

    @staticmethod
    def _read_file(filepath: str) -> Optional[bytes]:
        """Read xml data file from disk

        :param filepath: path to the data file
        :return: etree element
        """
        try:
            f = open(filepath, 'rb')
        except FileNotFoundError:
            logging.error(f'File not found: {filepath}')
            return

        try:
            data = f.read()
            logging.info(f'XML data file read: {filepath}')
            return data
        except ValueError:
            logging.error(f'Failed to read XML: {filepath}')
            data = None
        finally:
            f.close()

        return data


class JsonData:
    """Simple class representing json data

    Only very general methods are described here.
    Method to create, update or delete the data are at the record level.

    :ivar content: json content of the object

    :Example:

    >>> from almapiwrapper.record import JsonData
    >>> data = JsonData(filepath='path_to_some_json_file')
    >>> print(data)

    Will pretty print the json content.

    Data can be changed in the `content` attribute.
    """
    def __init__(self, content: Optional[Dict] = None, filepath: Optional[str] = None):
        """JsonData object constructor"""
        self.content = content
        if filepath is not None:
            self.content = self._read_file(filepath)

    def __str__(self):
        return json.dumps(self.content, indent=2)

    def __bytes__(self):
        data_str = json.dumps(self.content)
        return bytes(data_str, 'utf-8)')

    @staticmethod
    def _read_file(filepath: str) -> Optional[Dict]:
        """Read json data file from disk

        :param filepath: path to the data file
        :return: json data
        """
        try:
            f = open(filepath, 'rb')
        except FileNotFoundError:
            logging.error(f'File not found: {filepath}')
            return

        try:
            data = json.load(f)
            logging.info(f'Json data file read: {filepath}')
        except ValueError:
            logging.error(f'Failed to read json: {filepath}')
            data = None
        finally:
            f.close()

        return data


if __name__ == "__main__":
    pass



import unittest
import sys
import os

from almapiwrapper.config import RecSet, Job, Reminder, fetch_reminders
from almapiwrapper.record import JsonData, XmlData
from almapiwrapper import config_log

config_log("test.log")


class TestReminder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        if len(reminders) > 0:
            for r in reminders:
                r.delete()
        data = JsonData(filepath='test/data/reminder_test1.json')
        r = Reminder(zone='NZ', env='S', data=data, create_reminder=True)

    def test_get_reminder(self):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        self.assertEqual(reminders[0].reminder_type, 'Additional edition',
                         'Bad reminder type, should be "Additional edition"')
        self.assertEqual(reminders[0].status, 'NEW', 'Bad reminder status, should be "NEW"')


    def test_delete_reminder(self):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        reminders[0].delete()
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 0, 'Not able to delete the reminder')
        data = JsonData(filepath='test/data/reminder_test1.json')
        Reminder(zone='NZ', env='S', data=data, create_reminder=True)

    def test_update_reminder(self):
        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        reminders[0].data['text'] = 'Test update'
        reminders[0].update()

        reminders = fetch_reminders(zone='NZ', env='S', entity_id='991170687152205501')
        self.assertEqual(len(reminders), 1, 'Not able to fetch the reminder')
        self.assertEqual(reminders[0].entity_id, '991170687152205501', 'Reminder linked to the wrong record')
        self.assertEqual(reminders[0].data['text'], 'Test update', 'Reminder not updated')


class TestRecSet(unittest.TestCase):

    def test_fetch_members_of_user_set(self):

        # Fetch set data
        s = RecSet('15443075260005504', 'UBS', 'S')
        members = s.get_members()

        self.assertEqual(len(members), s.get_members_number(), 'Not able to fetch all members')
        self.assertEqual(s.get_content_type(), 'USER', f'Bad type of content, should be "USER" '
                                                       f'and is "{s.get_content_type()}"')

    def test_fetch_members_of_bib_set(self):

        # Fetch set data
        s = RecSet('15443075460005504', 'UBS', 'S')
        members = s.get_members()

        self.assertEqual(len(members), s.get_members_number(), 'Not able to fetch all members')
        self.assertEqual(s.get_content_type(), 'IEP', f'Bad type of content, should be "IEP" '
                                                      f'and is "{s.get_content_type()}"')

    @unittest.skip
    def test_run_job(self):

        parameters = XmlData(filepath='test/data/config_job_parameters1.xml')

        job = Job('44', 'UBS', 'S')
        result = job.run(parameters)
        self.assertIsNotNone(result, "No return value when starting the job")
        self.assertFalse(job.error, 'The job cannot be started')
        self.assertIsNotNone(job.instance_id, "A job ID doesn't exist")

        instance = job.get_instance_info()
        self.assertIsNotNone(instance, "A job instance cannot be retrieved")

    def test_monitor_job(self):
        job = Job('44', 'UBS', 'S')

        instance = job.get_instance_info('15442965370005504')
        self.assertFalse(job.error, 'Not able to fetch a job instance')

        state = job.check_instance_state('15442965370005504')
        self.assertEqual(state['progress'], 100, 'Impossible to get instance info')
        self.assertEqual(state['status'], 'COMPLETED_SUCCESS', 'Impossible to get instance info')

        instances = job.get_instances()
        self.assertTrue(instances.content['total_record_count'] > 20)


if __name__ == '__main__':
    unittest.main()



"""This module allow to create and manage sets in Alma"""
from typing import Literal, Optional, List, Dict, Union
import logging
from ..record import JsonData, Record, check_error
import requests


def fetch_reminders(zone: str,
                    env: Optional[Literal['P', 'S']] = 'P',
                    reminder_type: Optional[str] = None,
                    reminder_status: Optional[str] = None,
                    from_date: Optional[str] = None,
                    to_date: Optional[str] = None,
                    entity_id: Optional[str] = None) -> List['Reminder']:
    """This function fetch the data describing the reminders.

    :param zone: zone of the fee
    :param env: environment of the entity: 'P' for production and 'S' for sandbox
    :param reminder_type: filter by reminder type
    :param reminder_status: filter by reminder status
    :param from_date: filter by date from, format: YYYY-MM-DD
    :param to_date: filter by date to, format: YYYY-MM-DD
    :param entity_id: filter by entity id, for example mms_id for bib records
    :return: List of :class:`almapiwrapper.config.reminder.Reminder`
    """

    # Prepare params and filters
    params = {'limit': 100}
    if reminder_type is not None:
        params['type'] = reminder_type
    if reminder_status is not None:
        params['status'] = reminder_status
    if from_date is not None:
        params['from'] = from_date
    if to_date is not None:
        params['to'] = to_date
    if entity_id is not None:
        params['entity_id'] = entity_id

    reminders = []

    # Init counters
    offset = 0
    nb_total_records = 0

    # Handle offset if more than 100 results are available
    while offset == 0 or offset < nb_total_records:

        # Make request
        params['offset'] = offset

        r = requests.get(f'{Reminder.api_base_url}/conf/reminders',
                         params=params,
                         headers=Record.build_headers(data_format='json', env=env,
                                                      zone=zone, rights='RW', area='Conf'))

        # Check result
        if r.ok is True:
            reminders_list = JsonData(r.json())
            nb_total_records = int(reminders_list.content['total_record_count'])

            if 'reminder' in reminders_list.content and reminders_list.content['reminder'] is not None:
                reminders += [Reminder(reminder['id'], zone, env) for reminder in reminders_list.content['reminder']]
                nb_reminders = len(reminders_list.content['reminder'])
            else:
                nb_reminders = 0

            logging.info(f'fetch_reminders("{zone}", "{env}"): '
                         f'{offset + nb_reminders} / '
                         f'{nb_total_records} users data available')
            offset += 100

        else:
            _handle_error(r, 'unable to fetch reminder data', zone, env)

    return reminders


def _handle_error(r: requests.models.Response, msg: str, zone: str, env: str):
    """Set the record error attribute to True and write the logs about the error

    :param r: request response of the api
    :param msg: context message of the error
    :return: None
    """
    json_data = r.json()
    print(json_data)
    print(r.url)
    error_message = json_data['errorList']['error'][0]['errorMessage']

    logging.error(f'fetch_reminders({zone}, {env}) - {r.status_code}: '
                  f'{msg} / {error_message}')


class Reminder(Record):
    """Class representing a reminder
    """

    def __init__(self,
                 reminder_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 data: Optional[Union[Dict, JsonData]] = None,
                 create_reminder: Optional[bool] = False) -> None:
        """Constructor of `Reminder`

        :param reminder_id: reminder id
        :param zone: zone of the reminder
        :param env: environment of the reminder: 'P' for production and 'S' for sandbox
        :param data: data of the reminder, :class:`almapiwrapper.record.JsonData` object
        :param create_reminder: if True create the reminder in Alma
        """
        super().__init__(zone, env, data)
        self.area = 'Conf'
        self.format = 'json'
        self.reminder_id = reminder_id

        if data is not None:
            if not isinstance(data, JsonData):
                data = JsonData(data)
            self.data = data

        if create_reminder is True:
            self._create_reminder()

    def _fetch_data(self) -> Optional[JsonData]:
        """This method fetch the data describing the reminder.

        :return: :class:`almapiwrapper.record.JsonData` object
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/reminders/{self.reminder_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: reminder data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch reminder data')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.reminder_id}', '{self.zone}', '{self.env}')"

    @check_error
    def save(self) -> 'Reminder':
        """save() -> 'Reminder'
        Save a Reminder record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<entity_type>_<entity_id>/reminder_<IZ>_<reminder_id>_<version>.xml

        :return: object :class:`almapiwrapper.reminder.Reminder`
        """
        filepath = f'records/{self.zone}_{self.reminder_id}/reminder_{self.zone}_{self.reminder_id}.json'
        self._save_from_path(filepath)
        return self

    @check_error
    def update(self) -> 'Reminder':
        """update() -> 'Reminder'
        Update the reminder
        """
        r = self._api_call('put',
                           f'{self.api_base_url}/conf/reminders/{self.reminder_id}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok:
            logging.info(f'{repr(self)}: reminder updated.')
        else:
            self._handle_error(r, 'failed to update reminder')

        return self

    @check_error
    def delete(self) -> None:
        """delete() -> None
        Delete the reminder

        :return: None
        """
        r = self._api_call('delete',
                           f'{self.api_base_url}/conf/reminders/{self.reminder_id}',
                           headers=self._get_headers())

        if r.ok:
            logging.info(f'{repr(self)}: reminder deleted.')
        else:
            self._handle_error(r, 'failed to delete reminder')

    @check_error
    def get_reminder_id(self) -> str:
        """get_reminder_id() -> str
        Get the reminder id

        :return: string
        """
        return self.data['id']

    @check_error
    def _create_reminder(self) -> 'Reminder':
        """create() -> 'Reminder'
        Create the reminder

        :return: object :class:`almapiwrapper.reminder.Reminder`
        """
        r = self._api_call('post',
                           f'{self.api_base_url}/conf/reminders',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok:
            self.data = JsonData(r.json())
            self.reminder_id = self.get_reminder_id()
            logging.info(f'{repr(self)}: reminder created.')
        else:
            print(r.text)
            self._handle_error(r, 'failed to create reminder')

        return self

    @property
    @check_error
    def entity_id(self) -> str:
        """entity_id() -> str
        Get the entity id

        :return: string containing the entity id
        """
        return self.data['entity']['entity_id']

    @entity_id.setter
    @check_error
    def entity_id(self, entity_id: str) -> None:
        """entity_id() -> str
        Set the entity id

        :return: None
        """
        self.data['entity']['entity_id'] = entity_id

    @property
    @check_error
    def status(self) -> str:
        """status() -> str
        Get the status

        :return: string containing the status
        """
        return self.data['status']['value']

    @status.setter
    @check_error
    def status(self, status: str) -> None:
        """status() -> str
        Set the status

        :return: None
        """
        self.data['status']['value'] = status

    @property
    @check_error
    def reminder_type(self) -> str:
        """reminder_type() -> str
        Get the reminder type

        :return: string containing the reminder type
        """
        return self.data['type']['value']

    @reminder_type.setter
    @check_error
    def reminder_type(self, reminder_type: str) -> None:
        """reminder_type() -> str
        Set the reminder type

        :return: None
        """
        self.data['type']['value'] = reminder_type



"""This module allow to create and manage sets in Alma"""
from typing import Literal, Optional
import logging
from ..record import XmlData, Record, check_error
from ..inventory import IzBib, NzBib
from ..users import User


class RecSet(Record):
    """Class representing a set of records

    :ivar set_id: initial value: ID of the set
    :ivar zone: initial value: zone of the fee
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    """
    def __init__(self,
                 set_id: str,
                 zone: str,
                 env: Literal['P', 'S'] = 'P') -> None:
        """Constructor of `RecSet`
        """
        super().__init__(zone, env)
        self.area = 'Conf'
        self.format = 'xml'
        self.set_id = set_id
        self._members = None

    def _fetch_data(self) -> Optional[XmlData]:
        """This method fetch the data describing the set.

        :return: :class:`almapiwrapper.record.XmlData`
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/sets/{self.set_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: set data available')
            return XmlData(r.content)
        else:
            self._handle_error(r, 'unable to fetch set data')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.set_id}', '{self.zone}', '{self.env}')"

    @check_error
    def get_members_number(self) -> int:
        """get_members_number(self) -> int
        Return the number of members

        :return: int with the number of records

        .. note::
            If the record encountered an error, this
            method will be skipped."""
        return int(self.data.find('.//number_of_members').text)

    @check_error
    def get_set_type(self) -> str:
        """get_set_type(self) -> str
        Return the type of the set

        :return: str indicating the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        return self.data.find('.//type').text

    @check_error
    def get_content_type(self) -> str:
        """get_content_type(self) -> str
        Return the type of the set

        :return: str indicating the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        return self.data.find('.//content').text

    @check_error
    def delete(self) -> None:
        """Delete a set

        :return: None

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        r = self._api_call('delete',
                           f'{self.api_base_url}/conf/sets/{self.set_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: set deleted')
            return
        else:
            print(r.text)
            self._handle_error(r, f'unable to delete the set')

    @check_error
    def get_members(self) -> list[Record]:
        """get_members(self) -> list[Record]
        Return a list with all the records of the set

        :return: list of records depending on the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        # Prevent to fetch data twice, if already available
        if self._members is not None:
            return self._members

        self._members = []
        while len(self._members) < self.get_members_number():
            r = self._api_call('get',
                               f'{self.api_base_url}/conf/sets/{self.set_id}/members',
                               params={'limit': '100', 'offset': str(len(self._members))},
                               headers=self._get_headers())
            if r.ok is False:
                self._handle_error(r, f'{repr(self)}: unable to fetch set members')
                return self._members
            else:
                rec_ids = [rec_id.text for rec_id in XmlData(r.content).content.findall('.//member/id')]

                # Build members -> IZ mms_id
                if self.get_content_type() in ['BIB_MMS', 'IEP'] and self.zone != 'NZ':
                    self._members += [IzBib(rec_id, self.zone, self.env) for rec_id in rec_ids]

                # Build members -> IZ mms_id
                elif self.get_content_type() in ['BIB_MMS'] and self.zone == 'NZ':
                    self._members += [NzBib(rec_id, self.env) for rec_id in rec_ids]

                # Build members -> users
                elif self.get_content_type() == 'USER':
                    self._members += [User(rec_id, self.zone, self.env) for rec_id in rec_ids]

                # Type of content not integrated
                else:
                    self._members += rec_ids
                    logging.warning(f'{repr(self)}: "{self.get_content_type()}" not integrated, only id are returned.')

                logging.info(f'{repr(self)}: {len(self._members)} / {self.get_members_number()} members fetched')

        return self._members



import almapiwrapper.users as userslib
from typing import List, Optional, Literal, Union
from ..apikeys import ApiKeys
from ..record import Record, JsonData
import requests
import logging
import time


def fetch_users(q: str, zone: str, env: Optional[Literal['P', 'S']] = 'P') -> List[userslib.User]:
    """Search users in IZ according to a request

    :param q: request in API syntax. "user_group~06" will search every linked account
    :param zone: code of the IZ, NZ or "all" for all IZs NZ excepted.
    :param env: "P" for production, "S" for sandbox. Default is production
    :return: list of :class:`almapiwrapper.users.User`
    """
    error = False
    users = []
    list_iz = ApiKeys().get_iz_codes() if zone == 'all' else [zone]
    for z in list_iz:

        # Interrupt in case of any error
        if error is True:
            break

        # Init counters
        offset = 0
        nb_total_records = 0

        # Handle offset if more than 100 results are available
        while offset == 0 or offset < nb_total_records:

            # Make request
            r = requests.get(f'{userslib.User.api_base_url}',
                             params={'q': q, 'limit': 100, 'offset': offset},
                             headers=Record.build_headers(data_format='json', env=env,
                                                          zone=z, rights='RW', area='Users'))

            # Check result
            if r.ok is True:
                users_list = JsonData(r.json())
                nb_total_records = int(users_list.content['total_record_count'])

                if 'user' in users_list.content and users_list.content['user'] is not None:
                    users += [userslib.User(user['primary_id'], z, env) for user in users_list.content['user']]
                    nb_users = len(users_list.content['user'])
                else:
                    nb_users = 0

                logging.info(f'fetch_users("{q}", "{z}", "{env}"): '
                             f'{offset + nb_users} / '
                             f'{nb_total_records} users data available')
                offset += 100
            else:
                _handle_error(q, r, f'unable to fetch data', z, env)
                error = True
                break

    return users


def fetch_user_in_all_iz(primary_id: str, env: Optional[Literal['P', 'S']] = 'P') -> List[userslib.User]:
    """Fetch by primary ID a user in all IZ

    :param primary_id: primary ID of the user to search across all IZs
    :param env: "P" for production, "S" for sandbox. Default is production
    :return: list of :class:`almapiwrapper.users.User`
    """

    list_iz = ApiKeys().get_iz_codes()
    users = []
    for iz in list_iz:
        users_temp = fetch_users(f'primary_id~{primary_id}', iz, env)
        if len(users_temp) == 1:
            users += users_temp

    return users


def check_synchro(nz_users: Union[List[userslib.User], userslib.User]) -> Optional[List[userslib.User]]:
    """Test if a NZ user is synchronized with copies of the account across IZs.
    :param nz_users: list of :class:`almapiwrapper.users.User` or only one :class:`almapiwrapper.users.User`
    :return: list of :class:`almapiwrapper.users.User` with not synchronized IZ user accounts
    """

    not_synchro_iz_users = []
    if type(nz_users) is not list:
        nz_users = [nz_users]

    nb_users = len(nz_users)
    nz_users = [user for user in nz_users if user.zone == 'NZ']

    if len(nz_users) < nb_users:
        logging.error(f'Impossible to check synchronization on not NZ account')
        return None

    for nz_user in nz_users:
        nz_user.add_synchro_note()

    time.sleep(300)

    for nz_user in nz_users:
        nz_user.save()
        iz_users = fetch_user_in_all_iz(nz_user.primary_id, nz_user.env)

        for iz_user in iz_users:
            iz_user.save()
            if iz_user.check_synchro_note() is False:
                not_synchro_iz_users.append(iz_user)

    for nz_user in nz_users:
        nz_user.remove_synchro_note()

    if len(not_synchro_iz_users) > 0:
        logging.warning(f'Count of not synchronized users: {len(not_synchro_iz_users)}')
    else:
        logging.info(f'Count of not synchronized users: {len(not_synchro_iz_users)}')

    return not_synchro_iz_users


def _handle_error(q: str, r: requests.models.Response, msg: str, zone: str, env: str):
    """Set the record error attribute to True and write the logs about the error

    :param r: request response of the api
    :param msg: context message of the error
    :return: None
    """
    json_data = r.json()
    error_message = json_data['errorList']['error'][0]['errorMessage']

    logging.error(f'fetch_users({q}, {zone}, {env}) - {r.status_code}: '
                  f'{msg} / {error_message}')


if __name__ == "__main__":
    pass



import unittest
import sys
import os

from almapiwrapper.users import User, NewUser
from almapiwrapper.record import JsonData
from almapiwrapper import config_log

config_log("test.log")


class TestCreateUser(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        u = User('TestUser1', 'UBS', 'S')
        if u.data is not None:
            u.delete()

        u = User('TestUser2', 'UBS', 'S')
        if u.data is not None:
            u.delete()

        u = User('TestUser3', 'UBS', 'S')
        if u.data is not None:
            u.delete()

        u = User('TestUser5', 'UBS', 'S')
        if u.data is not None:
            u.delete()

    def test_create_user1(self):

        # Create new user
        data = JsonData(filepath='test/data/user_test1.json')
        u = NewUser('UBS', 'S', data)

        self.assertEqual(u.primary_id, 'TestUser1', 'Bad primary ID')
        self.assertIsNotNone(u.data, 'Unable to load json data')
        self.assertFalse(u.error, 'Error in record before creating the user')

        u = u.create()

        self.assertFalse(u.error, 'Error in record after creating the user')

        # Delete user
        u.delete()

    def test_duplicated_create_user2(self):

        # Create new user
        data = JsonData(filepath='test/data/user_test2.json')
        u1 = NewUser('UBS', 'S', data)
        u1 = u1.create()
        self.assertFalse(u1.error, 'duplicated record creation: error in first user')

        u2 = NewUser('UBS', 'S', data)
        u2 = u2.create()
        self.assertTrue(u2.error, 'duplicated record creation: no error in second user')

        u1.delete()
        self.assertFalse(u1.error, 'Error existing after deleting user')

    def test_duplicated_update_user3(self):

        # Create new user
        data = JsonData(filepath='test/data/user_test3.json')
        u1 = NewUser('UBS', 'S', data)
        u1 = u1.create()
        self.assertFalse(u1.error, 'error during user creation')

        u1 = User('TestUser3', 'UBS', 'S')
        u1.set_password('123new_pw123').update()
        self.assertFalse(u1.error, 'unable to update password')

        u1.delete()
        self.assertFalse(u1.error, 'Error existing after deleting user')

    def test_fetch_not_existing_record(self):
        # Fetch not existing user
        u = User('TestUser_not_existing', 'UBS', 'S')
        self.assertIsNone(u.data, 'Unable to fetch data with api of a not existing record')
        self.assertTrue(u.error, 'No error when fetching not existing user')

    def test_synchro_note(self):
        # Create new user
        data = JsonData(filepath='test/data/user_test5.json')
        u = NewUser('UBS', 'S', data).create()

        self.assertFalse(u.check_synchro_note(), 'User has already synchronization note')
        u.add_synchro_note()

        u = User('TestUser5', 'UBS', 'S')
        self.assertTrue(u.check_synchro_note(), 'User has no synchronization note')

        u.remove_synchro_note()

        u = User('TestUser5', 'UBS', 'S')
        self.assertFalse(u.check_synchro_note(), 'User has still a synchronization note')

        u.delete()

    @classmethod
    def tearDownClass(cls):
        u = User('TestUser', 'UBS', 'S')
        if u.data is not None:
            u.delete()


if __name__ == '__main__':
    unittest.main()
